from ruxit.api.base_plugin import RemoteBasePlugin
from ruxit.api.exceptions import ConfigException, AuthException

import logging
import json
import os
import socket
import datetime
import importlib
import time

logger = logging.getLogger(__name__)


class RemoteSAPHANADBPlugin(RemoteBasePlugin):

    def initialize(self, **kwargs):
        try:
            self.dbapi = importlib.import_module("hdbcli.dbapi")
        except ImportError as e:
            logger.exception(e)
            raise ConfigException("Unable to load hdbcli Library. See Documentation.")
        if str(self.config["host"]).strip() != '':
            self.host = str(self.config["host"]).strip()
        else:
            raise ConfigException("Host cannot be empty")

        self.customdevicename = str(self.config["customdevicename"]).strip()
        self.customgroupname = str(self.config["customgroupname"]).strip()

        if self.config["port"] != '':
            self.port = self.config["port"]
        else:
            raise ConfigException("Port cannot be empty")

        if str(self.config["auth_user"]).strip() != '':
            self.user = str(self.config["auth_user"]).strip()
        else:
            raise ConfigException("User cannot be empty")

        if str(self.config["auth_password"]).strip() != '':
            self.password = str(self.config["auth_password"]).strip()
        else:
            raise ConfigException("Password cannot be empty")

        if self.config['capture_slow']:
            self.capture_slow = True
        else:
            self.capture_slow = False

        if self.config['debug']:
            logger.setLevel(logging.DEBUG)
        else:
            logger.setLevel(logging.INFO)

    def query(self, **kwargs):
        try:
            start = time.perf_counter()
            connection = self.dbapi.connect(address=self.host, port=self.port, user=self.user, password=self.password,
                                            encrypt=True, sslValidateCertificate=False, connecttimeout=15000, communicationTimeout=10000)
            slow_connection = self.dbapi.connect(address=self.host, port=self.port, user=self.user, password=self.password,
                                                 encrypt=True, sslValidateCertificate=False, connecttimeout=15000, communicationTimeout=25000)
            self.createGroupAndDevice(connection)
            pluginVersion = kwargs["json_config"]["version"]
            self.gatherAndReportProperties(connection, pluginVersion)
            self.gatherMetrics(connection, slow_connection)
            self.reportMetrics()
            self.specialCaseLicenseMetric(connection)
            connection.close()
            slow_connection.close()
            end = time.perf_counter()
            logger.info(f"Execution took : {end-start:0.4f} seconds")
        except self.dbapi.Error as e:
            # See SAP Error Codes
            # https://help.sap.com/viewer/4fe29514fd584807ac9f2a04f6754767/2.0.03/en-US/20a78d3275191014b41bae7c4a46d835.html
            if str(e).startswith("(10,"):
                logger.exception("Authentication Issue: " + str(e))
                raise AuthException("Authentication Issue: " + str(e))
            elif str(e).startswith("(414,") or str(e).startswith("(415,") or str(e).startswith("(416,") or str(e).startswith("(663,"):
                logger.exception("Authorization Issue: " + str(e))
                raise AuthException("Authorization Issue: " + str(e))
            else:
                raise ConfigException("Could not connect to supplied Host and Port")
                logger.exception("Connection Issue: " + str(e))

    def createGroupAndDevice(self, connection):
        with connection.cursor() as cursor:
            try:
                cursor.execute("SELECT DATABASE_NAME FROM SYS.M_DATABASES;")
            except (self.dbapi.Error, self.dbapi.ProgrammingError) as e:
                # See SAP Error Codes (Link above)
                if str(e).startswith("(10,"):
                    logger.exception("Authentication Issue: " + str(e))
                    raise AuthException("Authentication Issue: " + str(e))
                elif str(e).startswith("(414,") or str(e).startswith("(415,") or str(e).startswith("(416,") or str(e).startswith("(663,"):
                    logger.exception("Authorization Issue: " + str(e))
                    raise AuthException("Authorization Issue: " + str(e))
                else:
                    logger.exception("Cannot get DB Name: " + str(e))
            dbNameList = list()
            try:
                for row in cursor:
                    dbNameList.append(row[0])
            except self.dbapi.ProgrammingError as e:
                logger.exception("Error parsing DB Name Results: " + str(e))
        try:
            dbName = dbNameList[0]
        except IndexError as e:
            logger.exception("Setting DB Name to DEFAULT because of error: " + str(e))
            dbName = str(self.port)
        if self.customgroupname != "":
            cdGroup = self.topology_builder.create_group(self.customgroupname, self.customgroupname)
        else:
            cdGroup = self.topology_builder.create_group("SAP HANADB", "SAP HANADB")
        if self.customdevicename != "":
            deviceName = self.customdevicename
        else:
            deviceName = self.host + " - " + dbName
        self.device = cdGroup.create_device(deviceName, deviceName)
        logger.debug("Custom Device: " + deviceName + " created or renewed.")
        if dbName != str(self.port):
            self.device.report_property("Database Name", dbName)

    def gatherAndReportProperties(self, connection, pluginVersion):
        logger.info("Gathering & Reporting Properties")
        self.device.report_property("Plugin version", pluginVersion)
        with connection.cursor() as cursor:
            # Find & Report License Memory Limit
            try:
                cursor.execute("SELECT PRODUCT_LIMIT FROM SYS.M_LICENSE;")
            except (self.dbapi.Error, self.dbapi.ProgrammingError) as e:
                logger.exception("Cannot get Licensed Memory Limit: " + str(e))
            try:
                for row in cursor:
                    licenseMemoryLimit = row[0]
                    self.device.report_property("Licensed Memory Limit (GB)", str(licenseMemoryLimit))
            except (self.dbapi.Error, self.dbapi.ProgrammingError) as e:
                logger.exception("Error in gathering Licensed Memory Limit: " + str(e))
            # Gather Connected Host incase it's different than endpoint used
            connectedHost = "%"
            try:
                connectedHostSQL = "SELECT HOST FROM SYS.M_CONNECTIONS WHERE CONNECTION_STATUS = 'RUNNING' AND USER_NAME = '" + self.user + "';"
                cursor.execute(connectedHostSQL)
            except (self.dbapi.Error, self.dbapi.ProgrammingError) as e:
                logger.exception("Cannot get Currently Connected to Host: " + str(e))
            try:
                for row in cursor:
                    connectedHost = row[0]
            except (self.dbapi.Error, self.dbapi.ProgrammingError) as e:
                logger.exception("Error in gathering Currently Connected to Host: " + str(e))
            try:
                cursor.execute("SELECT VALUE FROM SYS.M_HOST_INFORMATION WHERE key='build_version' AND HOST LIKE '" + connectedHost + "';")
            except (self.dbapi.Error, self.dbapi.ProgrammingError) as e:
                logger.exception("Cannot get DB Version: " + str(e))
            hanaVersion = ""
            try:
                for row in cursor:
                    hanaVersion = row[0]
            except (self.dbapi.Error, self.dbapi.ProgrammingError) as e:
                logger.exception("Error in gathering Version: " + str(e))
            self.device.report_property("Version", hanaVersion)
            logger.debug("HANA DB Version Set: " + hanaVersion)
            # Use SQL Query to get ports, use config & socket to get DNS Names & IP Address
            logger.debug("Setting Endpoints")
            # gather ports used on connected to host
            try:
                portSQL = "SELECT DISTINCT PORT, SQL_PORT FROM SYS.M_SERVICES WHERE (( SERVICE_NAME = 'indexserver' and COORDINATOR_TYPE  = 'MASTER') or SERVICE_NAME = 'xsengine') AND HOST LIKE '" + connectedHost + "';"
                cursor.execute(portSQL)
            except (self.dbapi.Error, self.dbapi.ProgrammingError) as e:
                logger.exception("Error in gathering Ports: " + str(e))
            portSet = set()
            try:
                for row in cursor:
                    for col in row:
                        if col != 0:
                            portSet.add(col)
            except (self.dbapi.Error, self.dbapi.ProgrammingError) as e:
                logger.exception("Cannot Parse Ports value from cursor response: " + str(e))
            if self.port not in portSet:
                portSet.add(self.port)
            portList = list(portSet)
            # gather IP address of Remote Extension Endpoint (self.host)
            ipAddress = ""
            try:
                ipAddress = socket.gethostbyname(self.host)
                dnsNamesSet = set()
                dnsNamesSet.add(self.host)
                fqdn = socket.getfqdn(ipAddress)
                dnsNamesSet.add(fqdn)
                dnsNamesList = list(dnsNamesSet)
            except socket.gaierror as e:
                logger.exception("Error with gathering HostName/IPAddress: " + str(e))
            logger.debug("DNS Names List: " + str(dnsNamesList))
            logger.debug("IP Address: " + str(ipAddress))
            logger.debug("Port List: " + str(portList))
            if (ipAddress != "" and portList and dnsNamesList):
                self.device.add_endpoint(ip=str(ipAddress), port_list=portList, dnsNames=dnsNamesList)
            # If Remote Extension Endpoint is different than host name
            # For example in a cloud scenario where host has public IP/DNS name and also private IP and name
            if self.host != connectedHost and connectedHost != '%':
                connectedDNSNamesList = list()
                connectedDNSNamesList.append(connectedHost)
                # Get IP address
                try:
                    cursor.execute("SELECT VALUE FROM SYS.M_HOST_INFORMATION WHERE key='net_ip_addresses' AND HOST LIKE '" + connectedHost + "';")
                except (self.dbapi.Error, self.dbapi.ProgrammingError) as e:
                    logger.exception("Cannot get Connected to Host IP addresses: " + str(e))
                connectedIPAddresses = None
                try:
                    for row in cursor:
                        connectedIPAddresses = row[0]
                except (self.dbapi.Error, self.dbapi.ProgrammingError) as e:
                    logger.exception("Error in gathering Connected to Host IP addresses: " + str(e))
                if connectedIPAddresses is not None:
                    splitConnectedIPAddress = connectedIPAddresses.split(",")
                    for ip in splitConnectedIPAddress:
                        logger.debug("Connected DNS Names List: " + str(connectedDNSNamesList))
                        logger.debug("Connected IP Address: " + str(ip))
                        logger.debug("Port List: " + str(portList))
                        self.device.add_endpoint(ip=str(ip), port_list=portList, dnsNames=connectedDNSNamesList)

    def gatherMetrics(self, connection, slow_connection):
        logger.info("Gathering Metrics from HANA DB")
        try:
            # metrics with queries are defined in metric.json file
            metricsFile = os.path.abspath(os.path.join(os.path.dirname(__file__),  "metrics.json"))
            with open(metricsFile, "r") as f:
                self.metriclist = json.loads(f.read())
            for metric in self.metriclist['metrics']:
                if metric['key'] == "databases.heapTemporaryMemoryUsedByHost" and self.capture_slow:
                    internal_connection = slow_connection
                    logger.debug(f"Using connection for slow queries for {metric['displayname']}")
                elif metric['key'] == "databases.heapTemporaryMemoryUsedByHost" and not self.capture_slow:
                    metric['reportMetric'] = False
                    logger.info(f"Not gathering or reporting slow query : {metric['displayname']}")
                    continue
                else:
                    internal_connection = connection
                with internal_connection.cursor() as cursor:
                    if metric['dimensions'] == "":
                        logger.debug("Gathering Metric: " + metric['displayname'])
                        logger.debug("    Query: " + metric['dbquery'])
                        try:
                            cursor.execute(metric['dbquery'])
                            results = cursor.fetchall()
                        except (self.dbapi.Error, self.dbapi.ProgrammingError) as e:
                            logger.exception("Error with query: " + str(e))
                        try:
                            if cursor.rowcount != 0:
                                for row in results:
                                    metric['value'] = row[0]
                                    logger.debug("Returned Value:" + str(metric['value']))
                            else:
                                metric['reportMetric'] = False  # if no rows are returned don't report metric
                                logger.debug("No Value in DB, Metric: " + str(metric['displayname']) + " will not be reported.")
                        except (self.dbapi.Error, self.dbapi.ProgrammingError) as e:
                            metric['reportMetric'] = False
                            logger.exception("Cannot parse value from cursor response. Will not report Metric: " + metric['displayname'])
                            logger.exception("Metric Above had error: " + str(e))
                    else:
                        logger.debug("Gathering Metric: " + metric['displayname'])
                        logger.debug("    Query: " + metric['dbquery'])
                        try:
                            cursor.execute(metric['dbquery'])
                            results = cursor.fetchall()
                        except (self.dbapi.Error, self.dbapi.ProgrammingError) as e:
                            logger.exception("Error with query: " + str(e))
                        try:
                            if cursor.rowcount != 0:
                                splitDimensions = metric['dimensions'].split(',')
                                numDimensions = len(splitDimensions)
                                metric['value'] = '{"values":['
                                for row in results:
                                    # Building custom JSON String
                                    tmpMetricStr = '{"dimension_name":"'
                                    for x in range(numDimensions):
                                        tmpMetricStr = tmpMetricStr + str(row[x]) + ','
                                    tmpMetricStr = tmpMetricStr[:-1]
                                    tmpMetricStr = tmpMetricStr + '","value":"' + str(row[numDimensions]) + '"},'
                                    metric['value'] = metric['value'] + tmpMetricStr
                                metric['value'] = metric['value'][:-1]
                                metric['value'] = metric['value'] + "]}"
                            else:
                                metric['reportMetric'] = False
                                logger.debug("No Value in DB, Metric: " + str(metric['displayname']) + " will not be reported.")
                        except (self.dbapi.Error, self.dbapi.ProgrammingError) as e:
                            metric['reportMetric'] = False
                            logger.exception("Cannot parse value from cursor response. Will not report Metric: " + metric['displayname'])
                            logger.exception("Metric Above had error: " + str(e))
        except ValueError as e:
            logger.exception(str(e))
        except OSError as e:
            logger.exception(str(e))

    def reportMetrics(self):
        # sends all collected metrics to Cluster
        logger.info("Reporting Metrics")
        for metric in self.metriclist['metrics']:
            if metric['reportMetric'] is not False:
                # if metric has no dimensions just send it straight
                logger.debug("Reporting Metric: " + str(metric['displayname']))
                logger.debug("    Metric Value: " + str(metric['value']))
                if metric['dimensions'] == "":
                    # send different metric types and make sure no values are 'None'
                    if metric['metricType'] == "absolute" and metric['value'] is not None:
                        self.device.absolute(key=metric['key'], value=metric['value'])
                    elif metric['metricType'] == "relative" and metric['value'] is not None:
                        self.device.relative(key=metric['key'], value=metric['value'])
                    elif metric['metricType'] == "relative" and metric['value'] is not None:
                        if self.verifyState(metric['states'], metric['value']):
                            self.device.state_metric(key=metric['key'], value=metric['value'])

                # for metrics with dimensions use value json to report
                elif metric['dimensions'] != "":
                    try:
                        metricValues = json.loads(metric['value'])
                        for value in metricValues['values']:
                            # build dimensions dictionary
                            splitGatheredDimensions = value['dimension_name'].split(",")
                            splitConfigDimensions = metric['dimensions'].split(",")
                            numGatheredDimensions = len(splitGatheredDimensions)
                            numConfigDimensions = len(splitConfigDimensions)
                            # make sure number of collected dimensions is equal to defined in metrics JSON
                            if numConfigDimensions == numGatheredDimensions:
                                dimensionsDict = dict()
                                for x in range(numGatheredDimensions):
                                    dimensionsDict[splitConfigDimensions[x].strip()] = splitGatheredDimensions[x].strip()
                                # send different metric types and make sure no values are 'None'
                                if metric['metricType'] == "absolute" and value['value'] is not None:
                                    self.device.absolute(key=metric['key'], value=float(value['value']), dimensions=dimensionsDict)
                                elif metric['metricType'] == "relative" and value['value'] is not None:
                                    self.device.relative(key=metric['key'], value=float(value['value']), dimensions=dimensionsDict)
                                elif metric['metricType'] == "state" and value['value'] is not None:
                                    if self.verifyState(metric['states'], value['value']):
                                        self.device.state_metric(key=metric['key'], value=value['value'], dimensions=dimensionsDict)
                            else:
                                logger.warn("CONFIG ERROR! Number of configured dimensions does not match gathered dimensions")
                                raise ConfigException("Extension Error - See Logs")
                    except json.decoder.JSONDecodeError as e:
                        logger.exception("Cannot decode JSON: " + str(e))
            else:
                logger.debug("Not Reporting Metric: " + str(metric['displayname']))

    def specialCaseLicenseMetric(self, connection):
        try:
            with connection.cursor() as cursor:
                cursor.execute("SELECT SYSTEM_ID, EXPIRATION_DATE, VALID FROM SYS.M_LICENSE;")
                results = cursor.fetchall()
                for row in results:
                    sysIDDict = dict()
                    sysIDDict["SYSTEM_ID"] = row[0]
                    expireDate = row[1]
                    valid = row[2]
                    if self.verifyState(["TRUE", "EXPIRING_SOON", "FALSE"], valid):
                        logger.debug("Reporting Metric: Valid License")
                        logger.debug("    Metric Value: " + str(sysIDDict) + " : Value: " + str(valid) + " : Expires : " + str(expireDate))
                        if valid == "FALSE":
                            self.device.state_metric(key="databases.validLicenseBySystemID", value="FALSE", dimensions=sysIDDict)
                        elif (valid == "TRUE") and (expireDate is not None):
                            now = datetime.datetime.now()
                            remaining = expireDate - now
                            logger.debug("License will expire in : " + str(remaining.days) + " Days")
                            if remaining.days < 15:  # Switch Licensing Valid to Expire soon at 15 days remaining
                                self.device.state_metric(key="databases.validLicenseBySystemID", value="EXPIRING_SOON", dimensions=sysIDDict)
                            else:
                                self.device.state_metric(key="databases.validLicenseBySystemID", value="TRUE", dimensions=sysIDDict)
                        else:
                            # True and never expires
                            self.device.state_metric(key="databases.validLicenseBySystemID", value="TRUE", dimensions=sysIDDict)
                    else:
                        logger.warning("Issue with License Validity")
        except (self.dbapi.Error, self.dbapi.ProgrammingError) as e:
            logger.exception("Error Gathering License Info: " + str(e))

    def verifyState(self, metricStates, value):
        if value.upper() in metricStates:
            return True
        else:
            logger.exception("State collected is not one of that is defined.")
            return False
